//*****************************************************************************
// Recurent.cpp: Implementation of the CRecurentNetwork class
//*****************************************************************************

#include "stdafx.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//*****************************************************************************
// Construction/Destruction
//*****************************************************************************

CRecurentNetwork::CRecurentNetwork(
	const int iSizeIn,
	const int iSizeRec,
	const int iSizeOut,
	CActivate &cActivate,
	const double dMomentum, /* = 0.0 */
	const double dAlpha /* = 0.0 */
)
:	// Limits used
	m_iSizeIn(iSizeIn),
	m_iSizeRec(iSizeRec),
	m_iSizeOut(iSizeOut),
	// Configuration
	dMomentum(dMomentum),
	dAlpha(dAlpha),
	// Weights
	m_cWeightsIn(iSizeRec, iSizeIn, -1.0, 1.0),
	m_cWeightsRec(iSizeRec, iSizeRec, -1.0, 1.0),
	m_cWeightsOut(iSizeOut, iSizeRec, -1.0, 1.0),
	m_cDiffWIn(iSizeRec, iSizeIn),
	m_cDiffWRec(iSizeRec, iSizeRec),
	m_cDiffWOut(iSizeOut, iSizeRec),
	// Recurent layer output
	m_cInputRec(iSizeRec),
	// Activation function
	m_cActivate(cActivate)
{
	Restart();
	puts(" - Input weights: ");
	m_cWeightsIn.Print(" % 6.4f");
	puts(" - Recurent weights: ");
	m_cWeightsRec.Print(" % 6.4f");
	puts(" - Output weights: ");
	m_cWeightsOut.Print(" % 6.4f");
	puts(" - Recurent input: ");
	m_cInputRec.Print(" % 6.4f");
	puts("");
}

//*****************************************************************************
// Methods/Operators
//*****************************************************************************

void CRecurentNetwork::Restart()
{
	int iIndex;

	for(iIndex = 0; iIndex < m_iSizeRec; iIndex ++)
	{
		m_cInputRec(iIndex) = 0.5;
	}
	m_cInputRec(m_iSizeRec - 1) = -1.0;
}

void CRecurentNetwork::CountForward(
	CVector &cInputIn, 
	CVector &cInputRec,
	CVector &cOutputRec,
	CVector &cOutputOut
)
{
	int iNeuron;
	int iInput;
	double dNet;
	
	// Dummy neuron :)
	cInputRec(m_iSizeRec - 1) = -1.0;

	// Output from hidden layer
	for(iNeuron = 0; iNeuron < m_iSizeRec; iNeuron++)
	{
		dNet = 0.0;
		for(iInput = 0; iInput < m_iSizeIn; iInput++)
			dNet += cInputIn(iInput) * m_cWeightsIn(iNeuron, iInput);
		for(iInput = 0; iInput < m_iSizeRec; iInput++)
			dNet += cInputRec(iInput) * m_cWeightsRec(iNeuron, iInput);
		cOutputRec(iNeuron) = m_cActivate.F(dNet);
	}

	// Dummy neuron :)
	cOutputRec(m_iSizeRec - 1) = -1.0;

	// Output from network
	for(iNeuron = 0; iNeuron < m_iSizeOut; iNeuron++)
	{
		dNet = 0.0;
		for(iInput = 0; iInput < m_iSizeRec; iInput++)
			dNet += cOutputRec(iInput) * m_cWeightsOut(iNeuron, iInput);
		cOutputOut(iNeuron) = m_cActivate.F(dNet);
	}
}

void CRecurentNetwork::CountBackward(
	CList &cInput,
	CVector &cDesired
)
{
	bool bWork;
	int iTime = cInput.GetLength();
	int iNeuron;
	int iInput;
	CVector *pInput = NULL;
	CVector *pRecurentIn = NULL;
	CVector *pRecurentOut = NULL;
	CList cRecurent;
	CVector cOutput(m_iSizeOut);
	CVector cDeltaOut(m_iSizeOut);
	CVector cDeltaRecIn(m_iSizeRec);
	CVector cDeltaRecOut(m_iSizeRec);
	CMatrix cSumDWRec(m_iSizeRec, m_iSizeRec);
	CMatrix cSumDWIn(m_iSizeRec, m_iSizeIn);

	if(iTime == 0)
		return;

	// 'Reset'
	pRecurentIn = new CVector(m_iSizeRec, 0.5);
	(*pRecurentIn)(m_iSizeRec - 1) = -1.0;
	cRecurent.Append(pRecurentIn);
	bWork = cInput.MoveFirst();

	// Count the output from the network
	while(bWork)
	{
		pInput = (CVector *)cInput.Item();

		pRecurentOut = new CVector(m_iSizeRec);
		CountForward(*pInput, *pRecurentIn, *pRecurentOut, cOutput);
		cRecurent.Append(pRecurentIn = pRecurentOut);

		bWork = cInput.MoveNext();
	}

	// Error propagation from output layer
	for(iNeuron = 0; iNeuron < m_iSizeOut; iNeuron++)
	{
		cDeltaOut(iNeuron) = (cDesired(iNeuron) - cOutput(iNeuron)) *
			m_cActivate.DiffF(cOutput(iNeuron));

		for(iInput = 0; iInput < m_iSizeRec; iInput++)
		{
			m_cDiffWOut(iNeuron, iInput) *= dMomentum; 
			m_cDiffWOut(iNeuron, iInput) += 
				dAlpha * cDeltaOut(iNeuron) * (*pRecurentOut)(iInput);
		}
	}

	cRecurent.MovePrev();
	pRecurentIn = (CVector *)cRecurent.Item();

	// Deltas for the recurent and input layer
	// Time t = T
	for(iNeuron = 0; iNeuron < m_iSizeRec - 1; iNeuron++)
	{
		// Delta T
		cDeltaRecIn(iNeuron) = 0.0;
		for(iInput = 0; iInput < m_iSizeOut; iInput++)
			cDeltaRecIn(iNeuron) += 
				cDeltaOut(iInput) * m_cWeightsOut(iInput, iNeuron);
		cDeltaRecIn(iNeuron) *= m_cActivate.DiffF((*pRecurentOut)(iNeuron));

		// Recurent weights
		for(iInput = 0; iInput < m_iSizeRec; iInput++)
			cSumDWRec(iNeuron, iInput) =
				dAlpha * cDeltaRecIn(iNeuron) * (*pRecurentIn)(iInput);

		// Input weights
		for(iInput = 0; iInput < m_iSizeIn; iInput++)
			cSumDWIn(iNeuron, iInput) =
				dAlpha * cDeltaRecIn(iNeuron) * (*pInput)(iInput);
	}

	// Time t = T-1 .. 1
	while(cInput.MovePrev() && cRecurent.MovePrev())
	{
		pInput = (CVector *)cInput.Item();
		pRecurentOut = pRecurentIn;
		pRecurentIn = (CVector *)cRecurent.Item();
		cDeltaRecOut = cDeltaRecIn;

		for(iNeuron = 0; iNeuron < m_iSizeRec - 1; iNeuron++)
		{
			// Delta t
			cDeltaRecIn(iNeuron) = 0.0;
			for(iInput = 0; iInput < m_iSizeRec; iInput++)
				cDeltaRecIn(iNeuron) += 
					cDeltaRecOut(iInput) * m_cWeightsRec(iInput, iNeuron);
			cDeltaRecIn(iNeuron) *= m_cActivate.DiffF((*pRecurentOut)(iNeuron));

			// Recurent weights
			for(iInput = 0; iInput < m_iSizeRec; iInput++)
				cSumDWRec(iNeuron, iInput) +=
					dAlpha * cDeltaRecIn(iNeuron) * (*pRecurentIn)(iInput);

			// Input weights
			for(iInput = 0; iInput < m_iSizeIn; iInput++)
				cSumDWIn(iNeuron, iInput) +=
					dAlpha * cDeltaRecIn(iNeuron) * (*pInput)(iInput);
		}
	}

	// Finaly update the output weights ...
	for(iNeuron = 0; iNeuron < m_iSizeOut; iNeuron++)
		for(iInput = 0; iInput < m_iSizeRec; iInput++)
			m_cWeightsOut(iNeuron, iInput) += m_cDiffWOut(iNeuron, iInput);

	// ... together with recurent and input weights
	for(iNeuron = 0; iNeuron < m_iSizeRec; iNeuron++)
	{
		for(iInput = 0; iInput < m_iSizeRec; iInput++)
		{
			m_cDiffWRec(iNeuron, iInput) *= dMomentum;
			m_cDiffWRec(iNeuron, iInput) += 
				(cSumDWRec(iNeuron, iInput) / iTime);
			m_cWeightsRec(iNeuron, iInput) += m_cDiffWRec(iNeuron, iInput);
		}

		for(iInput = 0; iInput < m_iSizeIn; iInput++)
		{
			m_cDiffWIn(iNeuron, iInput) *= dMomentum;
			m_cDiffWIn(iNeuron, iInput) += 
				(cSumDWIn(iNeuron, iInput) / iTime);
			m_cWeightsIn(iNeuron, iInput) += m_cDiffWIn(iNeuron, iInput);
		}
	}
}

void CRecurentNetwork::Process(
	CVector &cInput, 
	CVector &cOutput
)
{
	CVector cOutputRec(m_iSizeRec);
	
	CountForward(cInput, m_cInputRec, cOutputRec, cOutput);
	// Remember recurent output
	m_cInputRec = cOutputRec;
}

void CRecurentNetwork::Process(
	CList &cInput, 
	CList &cOutput
)
{
	CVector *pVectorIn = NULL;
	CVector *pVectorOut = NULL;
	bool bWork = cInput.MoveFirst();
	
	Restart();
	cOutput.Clear();

	while(bWork)
	{
		pVectorIn = (CVector *)cInput.Item();
		pVectorOut = new CVector(m_iSizeOut);
		Process(*pVectorIn, *pVectorOut);
		cOutput.Append(pVectorOut);

		bWork = cInput.MoveNext();
	}
}

/*
void CRecurentNetwork::Learn(
	CTrainSet &TrainSet,
	const double dErrorLimit,
	const double dTolerance
)
{
	// Indexes
	int iStage = 0;
	int iPattern;
	int iInput;
	// Pattern processed
	CPattern *pPattern = NULL;
	// Desired, computed output
	CVector *pDesired = NULL;
	CList cOutput;
	CVector *pOutput = NULL;
	// Cumulative error, ...
	double dError;
	double dDiff;
	double dNorm;
	// Number of correct classifications
	int iCorrect;
	int iEqual;

	do
	{
		iStage ++;


		TrainSet.Scramble();

		// Learning stage through all the patterns
		for(iPattern = 0; iPattern < TrainSet.GetSize(); iPattern++)
		{
			pDesired = (CVector *)TrainSet(iPattern).Desired.Item();
			CountBackward(TrainSet(iPattern).Input, *pDesired);
		}

		// Count the cumulative error from all the train patterns
		dError = 0.0;
		iCorrect = 0;
		for(iPattern = 0; iPattern < TrainSet.GetSize(); iPattern++)
		{
			Process(TrainSet(iPattern).Input, cOutput);

			dNorm = 0.0;
			iEqual = 1;
			cOutput.MoveLast();
			pOutput = (CVector *)cOutput.Item();
			pDesired = (CVector *)TrainSet(iPattern).Desired.Item();

			// Difference from desired
			for(iInput = 0; iInput < m_iSizeOut; iInput++)
			{
				dDiff = (*pDesired)(iInput) - (*pOutput)(iInput);

				// Check whether the output was correct
				if(fabs(dDiff) > dTolerance)
					iEqual = 0;

				dNorm += dDiff * dDiff;
			}
			// Update the error function & number of guest results
			dError += 0.5 * dNorm;
			iCorrect += iEqual;
		}
		
		// Show the progress
		fprintf(
			stderr, 
			"\rLearning stage: %5d  Error: %6.3f  #: %2d ",
			iStage, dError, iCorrect
		);

	} while(dError > dErrorLimit);

	// Show the progress
	fprintf(
		stderr, 
		"\rLearning stage: %5d  Error: %6.3f  #: %2d ",
		iStage, dError, iCorrect
	);

	printf("\nLearning finished after %d stages\n", iStage);
	printf(
		"Learning stage: %5d\n"
		"Cumulative error: %6.3f\n"
		"# of correct clasifications: %2d\n",
		iStage, dError, iCorrect
	);
}
*/
