//*****************************************************************************
// NeuralNetwork.cpp: Implementation of the CNeuralNetwork class
//*****************************************************************************

#include "stdafx.h"

//*****************************************************************************
// Construction/Destruction
//*****************************************************************************

CNeuralNetwork::CNeuralNetwork(
	const INT iSizeIn,
	const INT iSizeHid,
	const INT iSizeOut
)
:	// Weights matrices
	m_WeightsHid(iSizeHid, iSizeIn, (REAL) -1.0, (REAL) 1.0),
	m_WeightsOut(iSizeOut, iSizeHid, (REAL) -1.0, (REAL) 1.0),
	m_DeltaHid(iSizeHid, iSizeIn),
	m_DeltaOut(iSizeOut, iSizeHid),
	// Temporal output from hidden layer
	m_OutputHid(iSizeHid),
	m_OutputOut(iSizeOut),
	// Limits used
	m_iSizeIn(iSizeIn),
	m_iSizeHid(iSizeHid),
	m_iSizeOut(iSizeOut)
{
	puts(" - Hidden weights: ");
	m_WeightsHid.Print(" % 6.4f");
	puts("\n - Output weights: ");
	m_WeightsOut.Print(" % 6.4f");
	puts("");
}

//*****************************************************************************
// Methods/Operators
//*****************************************************************************

VOID CNeuralNetwork::Process(
	CVector &Inputs, 
	CVector &Outputs, 
	REAL rTolerance
)
{
	INT iNeuron, iInput;
	REAL rNet;
	
	// Ouptut from hidden layer
	for(iNeuron = (INT) 0; iNeuron < m_iSizeHid; iNeuron++)
	{
		rNet = (REAL) 0.0;
		for(iInput = (INT) 0; iInput < m_iSizeIn; iInput++)
			rNet += Inputs(iInput) * m_WeightsHid(iNeuron, iInput);
		m_OutputHid(iNeuron) = F(rNet);
	}
	// Dummy neuron :)
	m_OutputHid(m_iSizeHid - 1) = (REAL) -1.0;

	// Output from network
	for(iNeuron = (INT) 0; iNeuron < m_iSizeOut; iNeuron++)
	{
		rNet = (REAL) 0.0;
		for(iInput = (INT) 0; iInput < m_iSizeHid; iInput++)
			rNet += m_OutputHid(iInput) * m_WeightsOut(iNeuron, iInput);
		m_OutputOut(iNeuron) = F(rNet);
	}

	// Fit the output according to the tolerance
	for(iNeuron = (INT) 0; iNeuron < m_iSizeOut; iNeuron++)
	{
		if(m_OutputOut(iNeuron) < rTolerance)
			Outputs(iNeuron) = (REAL) 0.0;
		else if(m_OutputOut(iNeuron) > ((REAL) 1.0 - rTolerance))
			Outputs(iNeuron) = (REAL) 1.0;
		else
			// Not sure :)
			Outputs(iNeuron) = (REAL) 0.5;
	}
}

VOID CNeuralNetwork::Learn(
	const REAL rAlpha,
	const REAL rMomentum,
	CTrainSet &TrainSet,
	const REAL rTolerance,
	const REAL rPrecision,
	const INT iMaxStage,
	const BOOL bProgress
)
{
	// Indexes
	INT iPattern, iNeuron, iInput, iStage = (INT) 1;
	// Current output, output sigma, hidden sigma 
	CVector Outputs(m_iSizeOut), SigmaOut(m_iSizeOut), SigmaHid(m_iSizeHid);
	// Cumulative error, ...
	REAL rError, rDiff, rNorm;
	// Number of correct classifications
	INT iCorrect, iEqual;

	while(iStage <= iMaxStage)
	{
		TrainSet.Scramble();
		
		// Learning stage through all the patterns
		for(iPattern = (INT) 0; iPattern < TrainSet.GetSize(); iPattern++)
		{
			Process(TrainSet(iPattern).Inputs, Outputs, rTolerance);

			// Delta for network output
			for(iNeuron = (INT) 0; iNeuron < m_iSizeOut; iNeuron++)
				SigmaOut(iNeuron) = DiffF(m_OutputOut(iNeuron)) *
					(TrainSet(iPattern).Desired(iNeuron) - m_OutputOut(iNeuron));

			// Delta for the hidden layer
			for(iNeuron = (INT) 0; iNeuron < m_iSizeHid; iNeuron++)
			{
				SigmaHid(iNeuron) = (REAL) 0.0;
				for(iInput = (INT) 0; iInput < m_iSizeOut; iInput++)
					SigmaHid(iNeuron) += 
						SigmaOut(iInput) * m_WeightsOut(iInput, iNeuron);
				SigmaHid(iNeuron) *= DiffF(m_OutputHid(iNeuron));
			}

			// Update output weights
			for(iNeuron = (INT) 0; iNeuron < m_iSizeOut; iNeuron++)
				for(iInput = (INT) 0; iInput < m_iSizeHid; iInput++)
				{
					m_DeltaOut(iNeuron, iInput) *= rMomentum; 
					m_DeltaOut(iNeuron, iInput) += 
						rAlpha * SigmaOut(iNeuron) * m_OutputHid(iInput);
					m_WeightsOut(iNeuron, iInput) += m_DeltaOut(iNeuron, iInput);
				}
			
			// Update hidden weights
			for(iNeuron = (INT) 0; iNeuron < m_iSizeHid; iNeuron++)
				for(iInput = (INT) 0; iInput < m_iSizeIn; iInput++)
				{
					m_DeltaHid(iNeuron, iInput) *= rMomentum;
					m_DeltaHid(iNeuron, iInput) +=
						rAlpha * SigmaHid(iNeuron) * TrainSet(iPattern).Inputs(iInput);
					m_WeightsHid(iNeuron, iInput) += m_DeltaHid(iNeuron, iInput);
				}
		}

		// Count the cumulative error from all the train patterns
		rError = (REAL) 0.0;
		iCorrect = (INT) 0;
		for(iPattern = (INT) 0;	iPattern < TrainSet.GetSize(); iPattern++)
		{
			Process(TrainSet(iPattern).Inputs, Outputs, rTolerance);

			// Difference from desired
			rNorm = (REAL) 0.0;
			iEqual = (INT) 1;
			for(iNeuron = (INT) 0; iNeuron < m_iSizeOut; iNeuron++)
			{
				// Check whether the fixed output was correct
				if(TrainSet(iPattern).Desired(iNeuron) != Outputs(iNeuron))
					iEqual = (INT) 0;

				// Count error from continuos network output
				rDiff = TrainSet(iPattern).Desired(iNeuron) - m_OutputOut(iNeuron);
				rNorm += rDiff * rDiff;
			}
			// Update the error function & number of guest results
			rError += 0.5 * rNorm;
			iCorrect += iEqual;
		}
		
		if(iStage % 10 == 0)
		{
			// Show the progress
			if(bProgress)
				fprintf(
					stderr, 
					"\rLearning stage: %5d  Error: %6.3f  #: %2d ",
					iStage, rError, iCorrect
				);

			printf(
				"\nLearning stage: %d\n"
				" - Cumulative error: %f\n"
				" - # of correctly clasified patterns: %d\n",
				iStage, rError, iCorrect
			);
		}

		if(rPrecision * TrainSet.GetSize() < iCorrect)
		{
			iStage++;
			// Learning finished earlier
			break;
		}
		
		// Next stage
		iStage ++;
	}

	// Show the progress
	if(bProgress)
		fprintf(
			stderr, 
			"\rLearning stage: %5d  Error: %6.3f  #: %2d ",
			iStage - 1, rError, iCorrect
		);

	printf("\nLearning finished after %d stages\n", iStage - 1);
	printf(
		"\nLearning stage: %d\n"
		" - Cumulative error: %f\n"
		" - # of correctly clasified patterns: %d\n",
		iStage - 1, rError, iCorrect
	);

}

REAL CNeuralNetwork::F(REAL rNet)
{
	return 1.0 / (1.0 + exp(-rNet));
}

REAL CNeuralNetwork::DiffF(REAL rO)
{
	return rO * (1 - rO);
}
